# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Optional

from ..._models import BaseModel

__all__ = ["PropertyDetailsResponse", "Result", "ResultPropertyDetails"]


class ResultPropertyDetails(BaseModel):
    city: str
    """The city name where the property resides."""

    postal_code: str
    """The 5 digit zip or postal code of the location where the property resides."""

    state: str
    """The state abbreviation where the property resides (e.g. `"CA"`)."""

    street: str
    """
    The full street address where the property resides, including the street number
    and street name.
    """

    above_grade_sqft: Optional[int] = None
    """The living_area_sqft - basement_sqft or `null`"""

    basement_sqft: Optional[int] = None
    """The approximate square footage of the basement."""

    bathrooms_full: Optional[int] = None
    """The number of full baths, capped at 5.

    A full bath contains all 4 of the 4 elements constituting a bath: Toilet, Sink,
    Bathtub, or Shower Head. A Full Bath will typically contain four elements; Sink,
    Toilet, Tub, and Shower Head (in tub or stall).
    """

    bathrooms_half: Optional[int] = None
    """The number of half baths. A half bath contains a Toilet and a Sink"""

    bedrooms_total: Optional[int] = None
    """The number of bedrooms, capped at 5."""

    garage_spaces: Optional[int] = None
    """The number of spaces in the garage(s)."""

    has_private_pool: Optional[bool] = None
    """
    TRUE if the property has a privately owned pool that is included in the
    sale/lease.
    """

    is_btr: Optional[bool] = None
    """TRUE if the property is part of a build to rent.

    Build To Rent (BTR) is a type of planned community where the developer plans on
    renting all or a substantial section of a subdivision. Not set if unknown.
    """

    is_in_hoa: Optional[bool] = None
    """TRUE if the property is inside a hoa. Not set if unknown."""

    latitude: Optional[float] = None
    """The approximate latitude of the property."""

    living_area_sqft: Optional[int] = None
    """The approximate livable area within the structure, designated in Square Feet."""

    longitude: Optional[float] = None
    """The approximate longitude of the property."""

    lot_size_sqft: Optional[int] = None
    """The total square footage of the lot."""

    msa_name: Optional[str] = None
    """The property's Metropolitan Statistical Area (MSA) name.

    Not set if property is not in a Metropolitan Statistical Area.
    """

    num_exterior_stories: Optional[int] = None
    """The number of above ground stories on the property."""

    ownership_profile: Optional[str] = None
    """The property owner's estimated portfolio size.

    The output can be one of the following: "<100", "100-1k", "1k-20k", "20k+".
    """

    slug: Optional[str] = None
    """A URL-safe string identifying the property address.

    The slug may be used in Rental Advisor URLs (e.g.
    https://rentaladvisor.opendoor.com/address/{slug}).
    """

    structure_type: Optional[str] = None
    """The type of structure that the property completely or partially encompasses.

    For example, House or Cabin are the overall structure and typically sold or
    leased as a whole. Multi Family and Docks may be sold in whole, but are often
    sold or leased by unit/slip. This field is the type of structure as opposed to
    style, which is under the Architectural Style field.
    """

    subdivision_name: Optional[str] = None
    """The name of the subdivision where the property is located."""

    unit: Optional[str] = None
    """The unit type and number of the property (e.g. `"Apt 123"`)."""

    year_built: Optional[int] = None
    """The year the property was built, rounded to the nearest 5."""


class Result(BaseModel):
    token: Optional[str] = None
    """The user-supplied token submitted for the address."""

    error_message: Optional[str] = None
    """The error message if the system was unable to process this address."""

    has_errors: Optional[bool] = None
    """Boolean indicating if the system was unable to process this address."""

    property_details: Optional[ResultPropertyDetails] = None
    """The property details for the given address."""


class PropertyDetailsResponse(BaseModel):
    results: List[Result]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from .crime_data_response import CrimeDataResponse as CrimeDataResponse
from .rental_comps_response import RentalCompsResponse as RentalCompsResponse
from .crime_data_fetch_params import CrimeDataFetchParams as CrimeDataFetchParams
from .rent_estimates_response import RentEstimatesResponse as RentEstimatesResponse
from .school_ratings_response import SchoolRatingsResponse as SchoolRatingsResponse
from .property_values_response import PropertyValuesResponse as PropertyValuesResponse
from .property_details_response import PropertyDetailsResponse as PropertyDetailsResponse
from .rental_comps_fetch_params import RentalCompsFetchParams as RentalCompsFetchParams
from .rent_estimates_fetch_params import RentEstimatesFetchParams as RentEstimatesFetchParams
from .school_ratings_fetch_params import SchoolRatingsFetchParams as SchoolRatingsFetchParams
from .property_values_fetch_params import PropertyValuesFetchParams as PropertyValuesFetchParams
from .property_details_fetch_params import PropertyDetailsFetchParams as PropertyDetailsFetchParams



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Optional
from datetime import date
from typing_extensions import Literal

from ..._models import BaseModel

__all__ = [
    "RentalCompsResponse",
    "Result",
    "ResultRentalComps",
    "ResultRentalCompsPropertyDetails",
    "ResultSubjectPropertyDetails",
]


class ResultRentalCompsPropertyDetails(BaseModel):
    city: str
    """The city name where the property resides."""

    postal_code: str
    """The 5 digit zip or postal code of the location where the property resides."""

    state: str
    """The state abbreviation where the property resides (e.g. `"CA"`)."""

    street: str
    """
    The full street address where the property resides, including the street number
    and street name.
    """

    above_grade_sqft: Optional[int] = None
    """The living_area_sqft - basement_sqft or `null`"""

    basement_sqft: Optional[int] = None
    """The approximate square footage of the basement."""

    bathrooms_full: Optional[int] = None
    """The number of full baths, capped at 5.

    A full bath contains all 4 of the 4 elements constituting a bath: Toilet, Sink,
    Bathtub, or Shower Head. A Full Bath will typically contain four elements; Sink,
    Toilet, Tub, and Shower Head (in tub or stall).
    """

    bathrooms_half: Optional[int] = None
    """The number of half baths. A half bath contains a Toilet and a Sink"""

    bedrooms_total: Optional[int] = None
    """The number of bedrooms, capped at 5."""

    garage_spaces: Optional[int] = None
    """The number of spaces in the garage(s)."""

    has_private_pool: Optional[bool] = None
    """
    TRUE if the property has a privately owned pool that is included in the
    sale/lease.
    """

    is_btr: Optional[bool] = None
    """TRUE if the property is part of a build to rent.

    Build To Rent (BTR) is a type of planned community where the developer plans on
    renting all or a substantial section of a subdivision. Not set if unknown.
    """

    is_in_hoa: Optional[bool] = None
    """TRUE if the property is inside a hoa. Not set if unknown."""

    latitude: Optional[float] = None
    """The approximate latitude of the property."""

    living_area_sqft: Optional[int] = None
    """The approximate livable area within the structure, designated in Square Feet."""

    longitude: Optional[float] = None
    """The approximate longitude of the property."""

    lot_size_sqft: Optional[int] = None
    """The total square footage of the lot."""

    msa_name: Optional[str] = None
    """The property's Metropolitan Statistical Area (MSA) name.

    Not set if property is not in a Metropolitan Statistical Area.
    """

    num_exterior_stories: Optional[int] = None
    """The number of above ground stories on the property."""

    ownership_profile: Optional[str] = None
    """The property owner's estimated portfolio size.

    The output can be one of the following: "<100", "100-1k", "1k-20k", "20k+".
    """

    slug: Optional[str] = None
    """A URL-safe string identifying the property address.

    The slug may be used in Rental Advisor URLs (e.g.
    https://rentaladvisor.opendoor.com/address/{slug}).
    """

    structure_type: Optional[str] = None
    """The type of structure that the property completely or partially encompasses.

    For example, House or Cabin are the overall structure and typically sold or
    leased as a whole. Multi Family and Docks may be sold in whole, but are often
    sold or leased by unit/slip. This field is the type of structure as opposed to
    style, which is under the Architectural Style field.
    """

    subdivision_name: Optional[str] = None
    """The name of the subdivision where the property is located."""

    unit: Optional[str] = None
    """The unit type and number of the property (e.g. `"Apt 123"`)."""

    year_built: Optional[int] = None
    """The year the property was built, rounded to the nearest 5."""


class ResultRentalComps(BaseModel):
    close_price: Optional[int] = None
    """
    The approximate last known closed monthly rental price for the property,
    denominated in dollars and cents.
    """

    close_price_date: Optional[date] = None
    """
    The approximate earliest known calendar day on which the closed rent became
    effective.
    """

    concession_description: Optional[str] = None
    """Description of any concessions available for the property."""

    concession_end_date: Optional[date] = None
    """End date of the concession period."""

    concession_start_date: Optional[date] = None
    """Start date of the concession period."""

    distance_miles: Optional[float] = None
    """The distance in miles from the subject property to the comp."""

    dom: Optional[int] = None
    """Days on Market for this property."""

    initial_list_price: Optional[int] = None
    """The initial list price of the property."""

    initial_list_price_date: Optional[date] = None
    """
    The approximate first known date at which the listed rent was publicly displayed
    or updated.
    """

    last_event_date: Optional[date] = None
    """The date which the last listing event occurred.

    This usually means a change in `listing_status`.
    """

    last_list_price: Optional[int] = None
    """
    The approximate last known published list rent for the property, denominated in
    dollars and cents.
    """

    listing_description: Optional[str] = None
    """Complete listing description as mentioned on the listing source website."""

    listing_status: Optional[Literal["active", "removed", "closed"]] = None
    """
    The status of the property: `active` refers to properties that are currently
    available for
    rent."closed"`refers to properties that were previously available for rent but have since been rented out.`removed`
    indicates that the listing associated with this property is no longer active.
    """

    move_out_date: Optional[date] = None
    """
    The date when the tenant moved out from the leased property during the previous
    lease.
    """

    ownership_profile: Optional[str] = None

    property_details: Optional[ResultRentalCompsPropertyDetails] = None
    """The property details for the comp."""

    response_codes: Optional[List[str]] = None
    """A list containing information about this comp subject property pairing.

    The possible response codes are as following:

    - estimated_close_price: the comp's listing has closed, but the actual close
      price is not yet known. In this case, the last list price is used as an
      estimate. close_price will be equal to the last known list_price
    """

    similarity_score: Optional[float] = None
    """
    A score between 0 and 1 describing how similar this comp is to the subject
    property. Similarity accounts for both home details (bed, bath, year built,
    etc.) as well as location. A larger score indicates a more similar comp.
    """


class ResultSubjectPropertyDetails(BaseModel):
    city: str
    """The city name where the property resides."""

    postal_code: str
    """The 5 digit zip or postal code of the location where the property resides."""

    state: str
    """The state abbreviation where the property resides (e.g. `"CA"`)."""

    street: str
    """
    The full street address where the property resides, including the street number
    and street name.
    """

    above_grade_sqft: Optional[int] = None
    """The living_area_sqft - basement_sqft or `null`"""

    basement_sqft: Optional[int] = None
    """The approximate square footage of the basement."""

    bathrooms_full: Optional[int] = None
    """The number of full baths, capped at 5.

    A full bath contains all 4 of the 4 elements constituting a bath: Toilet, Sink,
    Bathtub, or Shower Head. A Full Bath will typically contain four elements; Sink,
    Toilet, Tub, and Shower Head (in tub or stall).
    """

    bathrooms_half: Optional[int] = None
    """The number of half baths. A half bath contains a Toilet and a Sink"""

    bedrooms_total: Optional[int] = None
    """The number of bedrooms, capped at 5."""

    garage_spaces: Optional[int] = None
    """The number of spaces in the garage(s)."""

    has_private_pool: Optional[bool] = None
    """
    TRUE if the property has a privately owned pool that is included in the
    sale/lease.
    """

    is_btr: Optional[bool] = None
    """TRUE if the property is part of a build to rent.

    Build To Rent (BTR) is a type of planned community where the developer plans on
    renting all or a substantial section of a subdivision. Not set if unknown.
    """

    is_in_hoa: Optional[bool] = None
    """TRUE if the property is inside a hoa. Not set if unknown."""

    latitude: Optional[float] = None
    """The approximate latitude of the property."""

    living_area_sqft: Optional[int] = None
    """The approximate livable area within the structure, designated in Square Feet."""

    longitude: Optional[float] = None
    """The approximate longitude of the property."""

    lot_size_sqft: Optional[int] = None
    """The total square footage of the lot."""

    msa_name: Optional[str] = None
    """The property's Metropolitan Statistical Area (MSA) name.

    Not set if property is not in a Metropolitan Statistical Area.
    """

    num_exterior_stories: Optional[int] = None
    """The number of above ground stories on the property."""

    ownership_profile: Optional[str] = None
    """The property owner's estimated portfolio size.

    The output can be one of the following: "<100", "100-1k", "1k-20k", "20k+".
    """

    slug: Optional[str] = None
    """A URL-safe string identifying the property address.

    The slug may be used in Rental Advisor URLs (e.g.
    https://rentaladvisor.opendoor.com/address/{slug}).
    """

    structure_type: Optional[str] = None
    """The type of structure that the property completely or partially encompasses.

    For example, House or Cabin are the overall structure and typically sold or
    leased as a whole. Multi Family and Docks may be sold in whole, but are often
    sold or leased by unit/slip. This field is the type of structure as opposed to
    style, which is under the Architectural Style field.
    """

    subdivision_name: Optional[str] = None
    """The name of the subdivision where the property is located."""

    unit: Optional[str] = None
    """The unit type and number of the property (e.g. `"Apt 123"`)."""

    year_built: Optional[int] = None
    """The year the property was built, rounded to the nearest 5."""


class Result(BaseModel):
    api_code: int
    """An HTTP status code indicating the result of the address request."""

    rental_comps: List[ResultRentalComps]
    """The list of rental comps for the address."""

    token: Optional[str] = None
    """The user-supplied token submitted for the address."""

    error_message: Optional[str] = None
    """The error message if the system was unable to process this address."""

    has_errors: Optional[bool] = None
    """Boolean indicating if the system was unable to process this address.

    DEPRECATED: Use `api_code` instead.
    """

    subject_property_details: Optional[ResultSubjectPropertyDetails] = None
    """The property details for the address."""


class RentalCompsResponse(BaseModel):
    results: List[Result]
    """The list of rental comps for each address."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Optional

from ..._models import BaseModel

__all__ = ["CrimeDataResponse", "Result"]


class Result(BaseModel):
    address_token: str

    token: Optional[str] = None
    """The user-supplied token submitted for the address."""

    county_percentile: Optional[float] = None
    """
    The percentile rank within the county for the number of crime incidents near the
    block of the property.
    """

    error_message: Optional[str] = None
    """The error message if the system was unable to process this address."""

    nation_percentile: Optional[float] = None
    """
    The percentile rank within the US for the number of crime incidents near the
    block of the property.
    """


class CrimeDataResponse(BaseModel):
    results: List[Result]
    """The list of crime data for each address."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import collections
import concurrent.futures
from typing import TYPE_CHECKING, Deque, Iterable

import httpx

from ..._types import NOT_GIVEN, Body, Query, Headers, NotGiven
from ..._utils import maybe_transform, async_maybe_transform
from ..._compat import cached_property
from ..._resource import SyncAPIResource, AsyncAPIResource
from ..._response import (
    to_raw_response_wrapper,
    to_streamed_response_wrapper,
    async_to_raw_response_wrapper,
    async_to_streamed_response_wrapper,
)
from ...types.data import rent_estimates_response, rent_estimates_fetch_params
from ..._base_client import make_request_options
from ...types.data.rent_estimates_response import RentEstimatesResponse

if TYPE_CHECKING:
    from ... import Mainstay

__all__ = ["RentEstimatesResource", "AsyncRentEstimatesResource"]


MAX_CONCURRENT_REQUESTS = 4
MAX_ADDRESSES_PER_RENT_ESTIMATES_REQUEST = 50
class RentEstimatesResource(SyncAPIResource):
    def __init__(self, client: Mainstay) -> None:
        super().__init__(client)
        self._executor = concurrent.futures.ThreadPoolExecutor(max_workers=MAX_CONCURRENT_REQUESTS)

    @cached_property
    def with_raw_response(self) -> RentEstimatesResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/mainstay-io/mainstay-python#accessing-raw-response-data-eg-headers
        """
        return RentEstimatesResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> RentEstimatesResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/mainstay-io/mainstay-python#with_streaming_response
        """
        return RentEstimatesResourceWithStreamingResponse(self)

    # noinspection DuplicatedCode
    def fetch(
        self,
        *,
        addresses: Iterable[rent_estimates_fetch_params.Address],
        max_addresses_per_request: int = MAX_ADDRESSES_PER_RENT_ESTIMATES_REQUEST,
        #
        # Stainless arguments.
        #
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> Iterable[rent_estimates_response.Result]:
        """
        Fetch rent estimates for addresses (custom code)

        Args:
          addresses: An array of address objects, each specifying a property location.

          max_addresses_per_request: The maximum number of addresses to include in each request.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        futures: Deque[concurrent.futures.Future[RentEstimatesResponse]] = collections.deque()
        address_iter = iter(addresses)

        def submit_request() -> None:
            chunked_addresses: list[rent_estimates_fetch_params.Address] = []
            for address in address_iter:
                chunked_addresses.append(address)
                if len(chunked_addresses) >= max_addresses_per_request:
                    break
            if chunked_addresses:
                futures.append(
                    self._executor.submit(
                        self._fetch,
                        addresses=chunked_addresses,
                        extra_headers=extra_headers,
                        extra_query=extra_query,
                        extra_body=extra_body,
                        timeout=timeout,
                    )
                )

        # Submit initial requests. MAX_CONCURRENT_REQUESTS will be submitted and also queue up.
        for _ in range(MAX_CONCURRENT_REQUESTS * 2):
            submit_request()

        while futures:
            future = futures.popleft()
            response = future.result()
            submit_request()

            yield from response.results

    def _fetch(
        self,
        *,
        addresses: Iterable[rent_estimates_fetch_params.Address],
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> RentEstimatesResponse:
        """
        Fetch rent estimates for addresses

        Args:
          addresses: An array of address objects, each specifying a property location.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._post(
            "/api/v2/data/rent-estimates",
            body=maybe_transform({"addresses": addresses}, rent_estimates_fetch_params.RentEstimatesFetchParams),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=RentEstimatesResponse,
        )


class AsyncRentEstimatesResource(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncRentEstimatesResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/mainstay-io/mainstay-python#accessing-raw-response-data-eg-headers
        """
        return AsyncRentEstimatesResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncRentEstimatesResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/mainstay-io/mainstay-python#with_streaming_response
        """
        return AsyncRentEstimatesResourceWithStreamingResponse(self)

    async def _fetch(
        self,
        *,
        addresses: Iterable[rent_estimates_fetch_params.Address],
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> RentEstimatesResponse:
        """
        Fetch rent estimates for addresses

        Args:
          addresses: An array of address objects, each specifying a property location.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return await self._post(
            "/api/v2/data/rent-estimates",
            body=await async_maybe_transform(
                {"addresses": addresses}, rent_estimates_fetch_params.RentEstimatesFetchParams
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=RentEstimatesResponse,
        )


class RentEstimatesResourceWithRawResponse:
    def __init__(self, rent_estimates: RentEstimatesResource) -> None:
        self._rent_estimates = rent_estimates

        self._fetch = to_raw_response_wrapper(
            rent_estimates._fetch,
        )


class AsyncRentEstimatesResourceWithRawResponse:
    def __init__(self, rent_estimates: AsyncRentEstimatesResource) -> None:
        self._rent_estimates = rent_estimates

        self._fetch = async_to_raw_response_wrapper(
            rent_estimates._fetch,
        )


class RentEstimatesResourceWithStreamingResponse:
    def __init__(self, rent_estimates: RentEstimatesResource) -> None:
        self._rent_estimates = rent_estimates

        self._fetch = to_streamed_response_wrapper(
            rent_estimates._fetch,
        )


class AsyncRentEstimatesResourceWithStreamingResponse:
    def __init__(self, rent_estimates: AsyncRentEstimatesResource) -> None:
        self._rent_estimates = rent_estimates

        self._fetch = async_to_streamed_response_wrapper(
            rent_estimates._fetch,
        )
